Imports System.Xml

' Represents an item in the XML document.
Public Class DocNode
    ' The XML document's node for this item.
    Public XmlDocNode As XmlNode

    ' The node's fully-qualified name.
    Public FullName As String

    ' The last part of the node's name.
    Public ShortName As String

    ' The node's type code:
    '   A       Assembly
    '   #ctor   Constructor
    '   T       Type
    '   F       Field
    '   P       Property
    '   M       Method
    '   E       Event
    Public TypeName As String
    Public TypeCode As String

    ' The node's scope code:
    '   Public
    '   Protected
    '   Private
    '   Unknown
    Public ScopeName As String

    ' The node's logical children.
    Public Children As New Collection

    ' Save the node's name and type code.
    Public Sub New(ByVal xml_node As XmlNode, ByVal new_FullName As String, ByVal new_TypeCode As String)
        ' Save the XML node.
        XmlDocNode = xml_node

        ' Save the full name.
        FullName = new_FullName

        ' Get the short name.
        If FullName.Contains(".") Then
            ' See if the name contains parentheses.
            If FullName.Contains("(") Then
                ' Use the last "." before the "(".
                ShortName = FullName.Substring(FullName.LastIndexOf(".", FullName.IndexOf("(")) + 1)
            Else
                ' No parentheses. Use the last ".".
                ShortName = FullName.Substring(FullName.LastIndexOf(".") + 1)
            End If
        Else
            ShortName = FullName
        End If

        ' Save the type and scope codes.
        If ShortName.Contains("#ctor") Then
            ' This is a constructor.
            TypeCode = "#ctor"
        Else
            TypeCode = new_TypeCode
        End If
        TypeName = TypeCodeToString(TypeCode)

        ' Get the scope code.
        Dim scope_node As XmlNode = Nothing
        If xml_node IsNot Nothing Then scope_node = xml_node.Item("scope")
        If scope_node Is Nothing Then
            ScopeName = ""
        Else
            ScopeName = scope_node.InnerText
        End If
    End Sub

    ' Add a child to this node's Children collection.
    Public Sub AddChild(ByVal new_child As DocNode)
        ' See if the new child has our full name.
        If new_child.FullName = Me.FullName Then
            ' The child has information that belongs in our node.
            ' Save the XML node and type code, and discard the new child.
            Me.XmlDocNode = new_child.XmlDocNode
            Me.TypeCode = new_child.TypeCode
            Me.TypeName = new_child.TypeName
            Me.ScopeName = new_child.ScopeName
        Else
            ' Get the next piece of the child's name.
            ' Remove the first part.
            Dim name_piece As String = new_child.FullName.Substring(Me.FullName.Length + 1)

            ' Remove the last part. This is anything that comes after a "."
            ' unless the "." is inside parentheses (as in a constructor's parameters).
            If name_piece.Contains(".") Then
                ' See if we have parentheses.
                If name_piece.Contains("(") Then
                    ' Remove the end if the "." comes before the "(".
                    If name_piece.IndexOf(".") < name_piece.IndexOf("(") Then
                        name_piece = name_piece.Substring(0, name_piece.IndexOf("."))
                    End If
                Else
                    ' No parentheses. Remove the trailing part.
                    name_piece = name_piece.Substring(0, name_piece.IndexOf("."))
                End If
            End If

            ' See if we have a child with this short name.
            If Children.Contains(name_piece) Then
                ' We have this child. Get it.
                Dim child_node As DocNode = DirectCast(Children.Item(name_piece), DocNode)

                ' Put the new child inside this child.
                child_node.AddChild(new_child)
            Else
                ' We don't have the needed child. Make it.
                Dim child_node As New DocNode(Nothing, Me.FullName & "." & name_piece, "?")
                Me.Children.Add(child_node, name_piece)

                ' Put the new child inside this child.
                child_node.AddChild(new_child)
            End If
        End If
    End Sub

    ' Display the node's data.
    Public Function PrefixOrder(ByVal make_link As Boolean, ByVal show_scope As Boolean, ByVal show_type As Boolean, ByVal show_full_name As Boolean, ByVal show_summary As Boolean, Optional ByVal indent As Integer = 0) As String
        Dim result As String = _
            Description(make_link, show_scope, show_type, _
                show_full_name, show_summary, indent) & vbCrLf

        For Each child_node As DocNode In Children
            result &= child_node.PrefixOrder(make_link, show_scope, show_type, show_full_name, show_summary, indent + 1)
        Next child_node
        Return result
    End Function

    ' Return an HTML string for this object's ancestor hierarchy.
    Public Function Ancestors() As String
        ' Make links for ancestors.
        Dim txt As String = ""

        ' See if our name contains parentheses.
        Dim fields() As String
        If Me.FullName.Contains("(") Then
            ' Don't split inside the parentheses.
            fields = Me.FullName.Substring(0, Me.FullName.IndexOf("(")).Split("."c)
        Else
            fields = Me.FullName.Split("."c)
        End If
        Dim indent As String = ""
        Dim full_name As String = ""
        For i As Integer = 0 To fields.Length - 2
            full_name &= fields(i)
            txt &= indent & _
                "<A HREF=""" & full_name & ".html"">" & fields(i) & "</A><BR>" & vbCrLf
            full_name &= "_"
            indent &= "&nbsp;&nbsp;&nbsp;&nbsp;"
        Next i

        ' Make an entry for this object.
        txt &= indent & _
            "<B>" & Me.ShortName & "</B><BR>" & vbCrLf

        Return txt
    End Function

    ' Return an HTML string for this object's revision history.
    Public Function History() As String
        Dim txt As String = "<P><TABLE BORDER=""1"" CELLPADDING=""2"">" & vbCrLf
        txt &= "  <TR><TH ALIGN=""Left"">Date</TH><TH ALIGN=""Left"">By</TH><TH ALIGN=""Left"">Description</TH></TR>" & vbCrLf

        ' Get the History node.
        Dim history_node As XmlNode = Nothing
        If Me.XmlDocNode IsNot Nothing Then
            history_node = Me.XmlDocNode.Item("History")
        End If
        If history_node Is Nothing Then
            txt &= "<TR><TD COLSPAN=""3"" ALIGN=""Center""><FONT SIZE=""+1"" COLOR=""#FF0000""><B>Missing</B></FONT></TD></TR>" & vbCrLf
            txt &= "</TABLE>" & vbCrLf
            Return txt
        End If

        ' Read the change entries.
        For Each child_node As XmlNode In history_node.ChildNodes
            txt &= "  <TR>"
            txt &= "<TD ALIGN=""Right"">" & child_node.Attributes("Date").Value & "</TD>"
            txt &= "<TD>" & child_node.Attributes("By").Value & "</TD>"
            txt &= "<TD>" & child_node.InnerText & "</TD>"
            txt &= "</TR>" & vbCrLf
        Next child_node

        txt &= "</TABLE>" & vbCrLf
        Return txt
    End Function

    ' Return a short description for this object.
    Private Function Description(ByVal make_link As Boolean, ByVal show_scope As Boolean, ByVal show_type As Boolean, ByVal show_full_name As Boolean, ByVal make_index_row As Boolean, Optional ByVal indent As Integer = 0) As String
        Dim txt As String = New String(ControlChars.Tab, indent)

        If make_link Then txt &= "<A HREF=""" & _
            Me.FullName.Replace(".", "_") & ".html"">"

        If show_scope Then
            If Me.ScopeName.Length > 0 Then txt &= Me.ScopeName & " "
        End If
        If show_type Then
            txt &= Me.TypeName & " "
        End If
        If show_full_name Then
            txt &= Me.FullName
        Else
            txt &= Me.ShortName
        End If

        If make_link Then txt &= "</A>"

        If make_index_row Then
            txt = "  <TR><TD>" & txt & "</TD><TD>" & Me.Summary().Trim() & "</TD></TR>" & vbCrLf
        End If

        Return txt
    End Function

#Region "Return Basic Sections"
    ' Return the object's summary.
    Private Function Summary() As String
        If Me.XmlDocNode Is Nothing Then Return ""
        Dim summary_node As XmlNode = Me.XmlDocNode.Item("summary")
        If summary_node Is Nothing Then Return ""
        Return ProcessXml(summary_node)
    End Function

    ' Return the object's parameters.
    Private Function Parameters() As String
        If Me.XmlDocNode Is Nothing Then Return ""

        ' Get all param nodes that are children of this node.
        Dim parameter_nodes As XmlNodeList = Me.XmlDocNode.SelectNodes("./param")
        If parameter_nodes.Count = 0 Then Return ""

        ' Process the parameter nodes.
        Dim txt As String = "<H2><A NAME=""parameters"">Parameters</A></H2>" & vbCrLf
        txt &= "<TABLE BORDER=""1"" CELLPADDING=""2"">" & vbCrLf
        txt &= "  <TR><TH ALIGN=""Left"">Parameter</TH><TH ALIGN=""Left"">Description</TH></TR>" & vbCrLf
        For Each parameter_node As XmlNode In parameter_nodes
            txt &= "  <TR>"
            txt &= "<TD>" & parameter_node.Attributes("name").Value & "</TD>"
            txt &= "<TD>" & ProcessXml(parameter_node) & "</TD>"
            txt &= "</TR>" & vbCrLf
        Next parameter_node
        txt &= "</TABLE>" & vbCrLf

        Return txt
    End Function

    ' Return the object's remarks.
    Private Function Remarks() As String
        If Me.XmlDocNode Is Nothing Then Return ""
        Dim remarks_node As XmlNode = Me.XmlDocNode.Item("remarks")
        If remarks_node Is Nothing Then Return ""
        Return ProcessXml(remarks_node)
    End Function

    ' Return a link to the object's Members page.
    Private Function MembersLink(ByVal output_dir As String) As String
        Dim txt As String = ""
        If Me.Children.Count > 0 Then
            ' Compose the menu entry.
            txt = "<A HREF=""" & Me.FullName.Replace(".", "_") & _
                "__members.html"">Members</A><BR>" & vbCrLf

            ' Make the members page.
            MakeMembersPage(output_dir)
        End If

        Return txt
    End Function

    ' Return the See Also section.
    Private Function SeeAlso() As String
        If Me.XmlDocNode Is Nothing Then Return ""

        ' Get all nodes named "seealso" below this one.
        Dim seealso_nodes As XmlNodeList = Me.XmlDocNode.SelectNodes(".//seealso")
        If seealso_nodes.Count = 0 Then Return ""

        ' Process the seealso nodes.
        Dim txt As String = "<H2><A NAME=""seealso"">See Also</A></H2>" & vbCrLf
        txt &= "  <UL>" & vbCrLf
        For Each seealso_node As XmlNode In seealso_nodes
            ' Get the node's URL.
            Dim link_url As String = MakeHtmlLink(seealso_node.Attributes("cref").Value)

            ' Get the node's text.
            Dim link_text As String = seealso_node.InnerText

            txt &= "    <LI><A HREF=""" & link_url & """>" & link_text & "</A>" & vbCrLf
        Next seealso_node
        txt &= "  </UL>" & vbCrLf

        Return txt
    End Function
#End Region ' Return Basic Sections

    ' Return a string describing the type code.
    Private Function TypeCodeToString(ByVal type_code As String) As String
        Select Case type_code
            Case "A"
                Return "Assembly"
            Case "#ctor"
                Return "Constructor"
            Case "T"
                Return "Type"
            Case "F"
                Return "Field"
            Case "P"
                Return "Property"
            Case "M"
                Return "Method"
            Case "E"
                Return "Event"
            Case Else
                Return "Unknown"
        End Select
    End Function

    ' Sort the node's children in these groups:
    '   A       Assembly
    '   #ctor   Constructor
    '   T       Type
    '   F       Field
    '   P       Property
    '   M       Method
    '   E       Event
    ' And alphabetically within each group.
    Public Sub SortChildren()
        ' Copy the children into a temporary array.
        Dim new_children(0 To Children.Count - 1) As DocNode
        For i As Integer = 0 To Children.Count - 1
            new_children(i) = DirectCast(Children.Item(i + 1), DocNode)
        Next i

        ' Sort the array.
        Dim doc_node_comparer As New DocNodeComparer()
        Array.Sort(Of DocNode)(new_children, doc_node_comparer)

        ' Copy the sorted children back into the Children collection.
        Children = New Collection
        For i As Integer = 0 To new_children.Length - 1
            Children.Add(new_children(i), new_children(i).ShortName)
        Next i

        ' Recursively sort the children.
        For Each child_node As DocNode In Children
            child_node.SortChildren()
        Next child_node
    End Sub

    Private Class DocNodeComparer
        Implements IComparer(Of DocNode)

        ' Determines which type codes come first.
        Private m_TypeCodeRanks As New Collection

        ' Determines which scope come first.
        Private m_ScopeRanks As New Collection

        ' Initialize the type code ranks.
        Public Sub New()
            m_TypeCodeRanks.Add("A-", "A")
            m_TypeCodeRanks.Add("B-", "#ctor")
            m_TypeCodeRanks.Add("C-", "T")
            m_TypeCodeRanks.Add("D-", "F")
            m_TypeCodeRanks.Add("E-", "P")
            m_TypeCodeRanks.Add("F-", "M")
            m_TypeCodeRanks.Add("G-", "E")
            m_TypeCodeRanks.Add("Z-", "?")

            m_ScopeRanks.Add("A-", "Public")
            m_ScopeRanks.Add("B-", "Protected")
            m_ScopeRanks.Add("C-", "Private")
            m_ScopeRanks.Add("Z-", "?")
        End Sub
        Public Function TypeCodeRank(ByVal type_code As String) As String
            If m_TypeCodeRanks.Contains(type_code) Then Return CStr(m_TypeCodeRanks(type_code))
            Return CStr(m_TypeCodeRanks("?"))
        End Function
        Public Function ScopeNameRank(ByVal scope_code As String) As String
            If m_ScopeRanks.Contains(scope_code) Then Return CStr(m_ScopeRanks(scope_code))
            Return CStr(m_ScopeRanks("?"))
        End Function

        ' Compare two DocNodes.
        Public Function Compare(ByVal x As DocNode, ByVal y As DocNode) As Integer Implements System.Collections.Generic.IComparer(Of DocNode).Compare
            Dim x_name As String = _
                ScopeNameRank(x.TypeCode) & _
                TypeCodeRank(x.TypeCode) & _
                x.ShortName
            Dim y_name As String = _
                ScopeNameRank(y.TypeCode) & _
                TypeCodeRank(y.TypeCode) & _
                y.ShortName
            Return String.Compare(x_name, y_name)
        End Function
    End Class

    ' Make a Web page describing this object.
    Public Sub MakeWebPages(ByVal output_dir As String)
        Dim txt As String = My.Resources.ObjectPageTemplate
        txt = txt.Replace("@TITLE@", "About " & Me.ShortName)
        txt = txt.Replace("@DESCRIPTION@", Me.Description(False, True, True, True, False))

        ' Insert the members section.
        txt = txt.Replace("@MEMBERSMENU@", Me.MembersLink(output_dir))

        ' Insert the summary.
        txt = txt.Replace("@SUMMARY@", Me.Summary())

        ' Insert the parameters.
        Dim parameters_text As String = Me.Parameters()
        txt = txt.Replace("@PARAMETERS@", parameters_text)

        ' If the Parameters section is not empty,
        ' then insert a local link to it.
        If parameters_text.Length > 0 Then
            txt = txt.Replace("@PARAMETERSMENU@", "  <A HREF=""#parameters"">Parameters</A>")
        Else
            txt = txt.Replace("@PARAMETERSMENU@", "")
        End If

        ' Insert the remarks.
        txt = txt.Replace("@REMARKS@", Me.Remarks())

        ' Insert the ancestor hierarchy.
        txt = txt.Replace("@ANCESTORS@", Me.Ancestors())

        ' Insert the revision history.
        txt = txt.Replace("@HISTORY@", Me.History())

        ' Insert the See Also section.
        Dim seealso_text As String = Me.SeeAlso()
        txt = txt.Replace("@SEEALSO@", seealso_text)

        ' If the See Also section is not empty,
        ' then insert a local link to it.
        If seealso_text.Length > 0 Then
            txt = txt.Replace("@SEEALSOMENU@", "  <A HREF=""#seealso"">See Also</A>")
        Else
            txt = txt.Replace("@SEEALSOMENU@", "")
        End If

        ' Save the result.
        My.Computer.FileSystem.WriteAllText(output_dir & _
            Me.FullName.Replace(".", "_") & ".html", _
            txt, False)

        ' Recursively make children.
        For Each child_node As DocNode In Children
            child_node.MakeWebPages(output_dir)
        Next child_node
    End Sub

    ' Return HTML text representing this node's contents.
    Private Function ProcessXml(ByVal xml_node As XmlNode) As String
        ' Process the children.
        Dim txt As String = ""
        For Each child_node As XmlNode In xml_node.ChildNodes
            Select Case child_node.Name
                Case "#text"    ' Just some text.
                    txt &= child_node.Value
                Case "para"     ' Insert a paragraph.
                    txt &= "<P>"
                Case "list"     ' Make a table.
                    txt &= MakeTable(child_node)
                Case "see"      ' Make an inline link.
                    txt &= "<A HREF=""" & MakeHtmlLink(child_node.Attributes("cref").Value) & """>"
                    txt &= ProcessXml(child_node)
                    txt &= "</A>"
                Case "c"        ' Format as inline code <TT>.
                    txt &= "<TT>"
                    txt &= ProcessXml(child_node)
                    txt &= "</TT>"
                Case "code"     ' Format as code <PRE>.
                    txt &= "<PRE>"
                    Dim code_text As String = ProcessXml(child_node)

                    ' Trim off leading carriage returns and line feeds.
                    Dim crlf() As Char = New Char() { _
                        Microsoft.VisualBasic.ControlChars.Cr, _
                        Microsoft.VisualBasic.ControlChars.Lf _
                    }
                    code_text = code_text.TrimStart(crlf)

                    ' Trim all trailing whitespace.
                    code_text = code_text.TrimEnd()

                    txt &= code_text
                    txt &= "</PRE>"
                Case "seealso"
                    ' These are not displayed here. Instead they
                    ' are are collected in the See Also section.
                Case Else
                    ' Handle other cases here.
            End Select
        Next child_node

        Return txt
    End Function

    ' Return HTML text representing a <table> node.
    Private Function MakeTable(ByVal table_node As XmlNode) As String
        Dim txt As String = "<P><TABLE BORDER=""1"" CELLPADDING=""2"">" & vbCrLf
        For Each child_node As XmlNode In table_node.ChildNodes
            Select Case child_node.Name
                Case "listheader"
                    txt &= "  " & MakeTableItem(child_node, "TH") & vbCrLf
                Case "item"
                    txt &= "  " & MakeTableItem(child_node, "TD") & vbCrLf
            End Select
        Next child_node
        txt &= "</TABLE><P>" & vbCrLf
        Return txt
    End Function

    ' Return HTML text representing a <term>...</term><description>...</description> table item node.
    Private Function MakeTableItem(ByVal table_node As XmlNode, ByVal item_tag As String) As String
        Dim open_tag As String = "<" & item_tag & ">"
        Dim close_tag As String = "</" & item_tag & ">"
        Dim txt As String = "<TR>"
        txt &= open_tag & ProcessXml(table_node.Item("term")) & close_tag
        txt &= open_tag & ProcessXml(table_node.Item("description")) & close_tag
        txt &= "</TR>"
        Return txt
    End Function

    ' Convert the XML style link into an appropriate HTML style link.
    Private Function MakeHtmlLink(ByVal link_url As String) As String
        If link_url.Contains(":") Then link_url = link_url.Substring(link_url.IndexOf(":") + 1)
        link_url = link_url.Replace(".", "_") & ".html"
        Return link_url
    End Function

    ' Return HTML text representing a Members page for this DocNode.
    Private Sub MakeMembersPage(ByVal output_dir As String)
        Dim txt As String = My.Resources.MembersPageTemplate
        txt = txt.Replace("@TITLE@", Me.ShortName & " Members")

        Dim top_menu As String = ""

        ' Process our children.
        Dim body As String = ""
        Dim old_scope As String = Nothing
        Dim old_type As String = Nothing
        Dim TABLE_END As String = "</TABLE>" & vbCrLf & vbCrLf
        For Each child_node As DocNode In Me.Children
            ' See if we're starting a new section.
            Dim new_scope As String = child_node.ScopeName
            Dim new_type As String = child_node.TypeName
            If (new_scope <> old_scope) OrElse (new_type <> old_type) Then
                old_scope = new_scope
                old_type = new_type

                ' End the previous table.
                body &= TABLE_END

                ' Start the new section.
                Dim section_name As String = new_scope & "_" & new_type
                body &= "<H2><A NAME=""" & section_name & """>" & _
                    new_scope & " " & new_type & "</A></H2>" & vbCrLf
                body &= "<TABLE BORDER=""1"" WIDTH=""100%"">" & vbCrLf
                body &= "  <TR><TH WIDTH=""25%"" ALIGN=""Left"">Name</TH><TH WIDTH=""75%"" ALIGN=""Left"">Description</TH></TR>" & vbCrLf

                top_menu &= "<A HREF=""#" & section_name & """>" & new_scope & " " & new_type & "</A>&nbsp;&nbsp;"
            End If

            ' Display this child's entry.
            body &= "  <TR><TD>" & _
                child_node.Description(True, False, False, False, False) & _
                "</TD><TD>" & child_node.Summary() & "</TD></TR>" & vbCrLf
        Next child_node

        ' Remove the initial </TABLE> and add a final one.
        body = body.Substring(TABLE_END.Length) & "</TABLE>"

        ' Insert the body.
        txt = txt.Replace("@BODY@", body)

        ' Insert the top menu.
        If top_menu.Length > 0 Then
            top_menu &= vbCrLf & "<HR COLOR=""#0000C0"">" & vbCrLf
        End If
        txt = txt.Replace("@TOPMENU@", top_menu)

        ' Write the results.
        My.Computer.FileSystem.WriteAllText(output_dir & _
            Me.FullName.Replace(".", "_") & "__members.html", _
            txt, False)
    End Sub
End Class
